#ifndef bqtExprHH
#define bqtExprHH

//#define BINDING_DEBUG

#include <string>
#include <vector>

#include "autoptr"

typedef autoptr<class expr> exprp;

#include "tristate"
#include "tokens.hh"
#include "symtables.hh"
#include "lexer.hh"

enum ExprCompileDefs
{
    exprdef_nothingparticular,
    exprdef_may_produce_char_from_string,
    exprdef_cast_to_int,
    exprdef_membername,
    exprdef_pure_condition
};

class expr_visitor
{
public:
    virtual ~expr_visitor() {}
public:
    virtual void set_func_call(const std::string& name, unsigned param_index) = 0;
    virtual void expr_visit(exprp e, tristate write_target = false) = 0;
};
class expr: public ptrable
{
public:
    expr();
    virtual ~expr();
    
    tristate is_num() const;
    tristate is_string() const;
    virtual TypeP GetType() const = 0;
    virtual bool is_const() const { return false; }

    virtual bool MayErr() const { return true; }
    virtual bool IsComparison() const { return false; }

    virtual exprp optimized()
    {
        return this;
    }

    virtual void Compile(ExprCompileDefs defs=exprdef_nothingparticular);

    virtual void ForEachSubExpr(expr_visitor& ) {  }
};

class expr_string: public expr
{
    std::string value;
public:
    expr_string(): value() { }
    expr_string(const char* v) : value(v) { }
    expr_string(const std::string& v) : value(v) { }

    virtual TypeP GetType() const { return TypeStr; }
    virtual bool is_const() const  { return true; }
    
    const std::string& get() const { return value; }

    virtual void Compile(ExprCompileDefs defs);
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(expr_visitor& func);
};

class expr_number: public expr
{
    typedef long long int itype;
    typedef long double ftype;
    TypeP type;
    itype int_val;
    ftype float_val;
public:
    expr_number() : type(TypeInt), int_val(0), float_val(0) { }
    expr_number(long long int i) : type(), int_val(), float_val() { InitIntImpl(i); }
    expr_number(long int i)      : type(), int_val(), float_val() { InitIntImpl(i); }
    expr_number(int i)           : type(), int_val(), float_val() { InitIntImpl(i); }
    expr_number(long long int i, TypeP t) : type(t),int_val(i),float_val(i) {}
    expr_number(long int i, TypeP t)      : type(t),int_val(i),float_val(i) {}
    expr_number(int i, TypeP t)           : type(t),int_val(i),float_val(i) {}
    expr_number(long double f, TypeP t=TypeDbl)   : type(t),int_val((itype)f),float_val(f) {}
    expr_number(double f, TypeP t=TypeDbl)        : type(t),int_val((itype)f),float_val(f) {}

    virtual TypeP GetType() const { return type; }
    virtual bool is_const() const  { return true; }

    bool is_int() const;
    long double   get_f() const;
    long long int get_i() const;

private:
    void InitIntImpl(long long int i)
    {
        int_val = i;
        type = (i < -32768 || i > 32767) ? TypeLng : TypeInt;
    }

public:
    virtual void Compile(ExprCompileDefs defs);
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(expr_visitor& func);
};


class expr_unary: public expr
{
public:
    QBtoken token;
    exprp p;
public:
	expr_unary(QBtoken tok, exprp a) : token(tok), p(a) { }

    virtual TypeP GetType() const;

    virtual exprp optimized();

    virtual void Compile(ExprCompileDefs defs);
    virtual bool MayErr() const { return false; }

    virtual bool IsComparison() const { return token==NOT && p->IsComparison(); }

    virtual void ForEachSubExpr(expr_visitor& func);
};

class expr_binary: public expr
{
public:
    QBtoken token; 
    exprp p0, p1;
public:
	expr_binary(QBtoken tok, exprp a, exprp b) : token(tok), p0(a), p1(b) { }
	
    virtual TypeP GetType() const;
    /* Note: IMP and EQV have been eliminated in the parse phase. */

    virtual exprp optimized();

    virtual void Compile(ExprCompileDefs defs);

    virtual bool IsComparison() const;

    virtual void ForEachSubExpr(expr_visitor& func);
};

class expr_label: public expr
{
    long double label;
public:
    expr_label() : label(0) { }
    expr_label(long double f) : label(f) { }
    expr_label(long long int i) : label(i) { }

    virtual TypeP GetType() const { return 0; }
    
    long double get_label() const { return label; }

    virtual void Compile(ExprCompileDefs defs);
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(expr_visitor& func);
};

class expr_userfun: public expr
{
    std::string name;
    TypeP type;
    std::vector<exprp> params;
public:
    expr_userfun(const std::string& n) : name(n), type(), params() { }
    expr_userfun(const std::string& n,
                 TypeP t,
                 const std::vector<exprp>& p) : name(n), type(t), params(p) { }
    
    const std::string& getname() const { return name; }
    
    virtual TypeP GetType() const { return type; }
    
    friend class expr_var;

    virtual void Compile(ExprCompileDefs defs);

    virtual void ForEachSubExpr(expr_visitor& func);
};

class expr_builtin_numeric: public expr
{
    QBtoken token;
    std::vector<exprp> params;
    std::vector<bool> lvaluemap;
    std::string pattern;
public:
    expr_builtin_numeric(const LexLocation&, QBtoken tok, const std::string& p="")
        : token(tok), params(), lvaluemap(), pattern(p) { }
    expr_builtin_numeric(const LexLocation&, QBtoken tok, exprp a)
        : token(tok), params(), lvaluemap(), pattern()
        { params.push_back(a); }

    expr_builtin_numeric(QBtoken tok, const std::string& p="")
        : token(tok), params(), lvaluemap(), pattern(p) { }
    expr_builtin_numeric(QBtoken tok, exprp a)
        : token(tok), params(), lvaluemap(), pattern()
        { params.push_back(a); }

    void set_params(const std::vector<exprp>& p) { params=p; }
    void set_lvalue_map(const std::vector<bool>& l) { lvaluemap=l; }
    void set_pattern(const std::string& p) { pattern=p; }
    
    virtual TypeP GetType() const;

    virtual exprp optimized();

    virtual void Compile(ExprCompileDefs defs);

    virtual void ForEachSubExpr(expr_visitor& func);
};

class expr_builtin_string: public expr
{
    QBtoken token;
    std::vector<exprp> params;
    std::vector<bool> lvaluemap;
    std::string pattern;
public:
    expr_builtin_string(const LexLocation&, QBtoken tok, const std::string& p="")
        : token(tok), params(), lvaluemap(), pattern(p) { }
    expr_builtin_string(const LexLocation&, QBtoken tok, exprp a)
        : token(tok), params(), lvaluemap(), pattern()
      { params.push_back(a); }

    expr_builtin_string(QBtoken tok, const std::string& p="")
        : token(tok), params(), lvaluemap(), pattern(p) { }
    expr_builtin_string(QBtoken tok, exprp a)
        : token(tok), params(), lvaluemap(), pattern()
      { params.push_back(a); }

    void set_params(const std::vector<exprp>& p) { params=p; }
    void set_lvalue_map(const std::vector<bool>& l) { lvaluemap=l; }
    void set_pattern(const std::string& p) { pattern=p; }
    
    QBtoken GetToken() const { return token; }
    exprp   GetParam(int n=0) const { return n < params.size() ? params[n] : (exprp)0; }
    exprp&  GetParamRef(int n=0) { return params[n]; }
    
    virtual TypeP GetType() const { return TypeStr; }

    virtual exprp optimized();

    virtual void Compile(ExprCompileDefs defs);

    virtual void ForEachSubExpr(expr_visitor& func);
};

/* These can only occur in select-case */
class expr_caseelse: public expr
{
public:
    expr_caseelse() { }

    virtual TypeP GetType() const { return 0; }

    virtual bool is_const() const { return true; }

    virtual void Compile(ExprCompileDefs defs);

    virtual void ForEachSubExpr(expr_visitor& func);
};
class expr_caseis: public expr
{
public:
    QBtoken token; // relational comparison operator
    exprp ref;
public:
    expr_caseis(QBtoken t, exprp r) : token(t), ref(r){ }

    virtual TypeP GetType() const { return 0; }

    virtual bool is_const() const { return ref->is_const(); }

    virtual void Compile(ExprCompileDefs defs);

    virtual void ForEachSubExpr(expr_visitor& func);
};
class expr_caserange: public expr
{
public:
    exprp min, max;
public:
    expr_caserange(exprp mi, exprp ma) : min(mi), max(ma) { }

    virtual TypeP GetType() const { return 0; }

    virtual bool is_const() const { return min->is_const() && max->is_const(); }

    virtual void Compile(ExprCompileDefs defs);

    virtual void ForEachSubExpr(expr_visitor& func);
};
class expr_caseor: public expr
{
public:
    exprp a,b;
public:
    expr_caseor(exprp aa, exprp bb) : a(aa),b(bb) { }

    virtual TypeP GetType() const { return 0; }

    virtual bool is_const() const { return a->is_const() && b->is_const(); }

    virtual void Compile(ExprCompileDefs defs);

    virtual void ForEachSubExpr(expr_visitor& func);
};


class expr_var: public expr
{
    std::string name;
    std::vector<exprp> indexes;
    bool array_as_whole;
    autoptr<expr_var> member;            // NULL if none
    TypeP type;
public:
    expr_var(const std::string& n)
        : name(n), indexes(), array_as_whole(false),
          member(), type() { }
    
    virtual TypeP GetType() const { return type; }
    virtual bool is_const() const { return false; }
    
    void addindex(exprp p) { indexes.push_back(p); }
    void addmember(autoptr<expr_var> m) { member = m; }
    
    bool hasmember() const { return member; }
    bool hasindexes() const { return !indexes.empty(); }
    unsigned numindexes() const { return indexes.size(); }
    bool iswholearray() const { return array_as_whole; }
    void setwholearray(bool b=true) { array_as_whole = b; }
    
    bool isarray() const { return iswholearray() || hasindexes(); }
    
    const std::string& getname() const { return name; }
    const autoptr<expr_var>& getmember() const { return member; }
    const exprp& getindex(unsigned n) const { return indexes[n]; }
    
    void set_type(TypeP t) { type=t; }
    
    autoptr<expr_userfun> convert_to_functioncall(TypeP type) const;

    virtual void Compile(ExprCompileDefs defs);
    virtual bool MayErr() const
    {
        if(type && (
          type->arraytype == TypeInfo::DynamicArray
       || type->arraytype == TypeInfo::ParamArray
         )) return true;
        
        for(unsigned a=0; a<indexes.size(); ++a)
            if(indexes[a]->MayErr()) return true;
        return false;
    }

    virtual void ForEachSubExpr(expr_visitor& func);
};

class expr_returnvalueholder: public expr_var
{
public :
    explicit expr_returnvalueholder(const std::string& n)
        : expr_var(n) { }

    virtual void Compile(ExprCompileDefs defs);
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubExpr(expr_visitor& func);
};

class expr_lubound: public expr
{
    QBtoken token;
    exprp var;
    exprp num_index;
public:
    expr_lubound(QBtoken t, exprp a, exprp b = new expr_number(1))
        : token(t), var(a), num_index(b) { }

    virtual TypeP GetType() const { return TypeLng; }
    virtual bool is_const() const
    {
        if(!num_index->is_const()) return false;
        // Return true if "name" is a $STATIC array, false otherwise.
        return false;
    }

    virtual void Compile(ExprCompileDefs defs);

    virtual void ForEachSubExpr(expr_visitor& func);
};


#endif
