#include <cstdio>
#include <cstring>

#include "stream.hh"
#include "error.hh"
#include "input.hh"
#include "gfx.hh" 

#include <map>

/* File writing */
class QbFPstream: public PrintStreamBase
{
    FILE* fp;
    int xpos;
public:
    QbFPstream(int fd) : fp(fdopen(fd, "wb")), xpos(0) { }
    
    virtual unsigned GetXpos() const
    {
        return xpos;
    }
    virtual unsigned GetXsize() const
    {
        return 65535;
    }
    virtual void DoCRLF()
    {
        fputc('\n', fp);
    }
    virtual void Write(const void* buf, size_t n)
    {
        size_t res = fwrite(buf, 1, n, fp);
        // FIXME: error checking
        fflush(fp);
    }
};

/* File input */
class QbFIstream: public InputImplement
{
    FILE* fp;
public:
    QbFIstream(int fd) : fp(fdopen(fd, "rb")) { }

    virtual void ReadLine(BAS_STR& target)
    {
    }

    virtual void BeginInput()
    {
        InputError = false;
        // File:
        //   Do nothing. InputTo reads file.
    }

    virtual bool EndInputOkp()
    {
        // File:
        //   Return false if InputTo failed, true otherwise.

        if(InputError) return false;
    }

    virtual BAS_STR Input_N(size_t n)
    {
        BAS_STR buf(n, '\0');
        size_t res = fread(&buf[0], 1, n, fp);
        if(res == 0)
        {
            if(ferror(fp)) { clearerr(fp); B_ERROR_Errno(); }
        }
        buf.erase(buf.begin() + res, buf.end());
        return buf;
    }

    virtual int InputChar()
    {
    }

    virtual void GotOverflow()
    {
        B_ERROR(6); // overflow
        InputError = true;
    }
};

/* Console I/O */
class QbConsoleStream: public PrintStreamBase, public InputImplement
{
    BAS_STR InputLineBuffer;
    unsigned InputLineBufferPos;
public:
    QbConsoleStream() : PrintStreamBase(), InputImplement() { }
    
    virtual unsigned GetXpos() const
    {
        return B_FuncPOS();
    }
    virtual unsigned GetXsize() const
    {
        return B_FuncWIDTH();
    }
    virtual void DoCRLF()
    {
        GFX_OutS("\r", 1);
    }
    virtual void Write(const void* buf, size_t n)
    {
        GFX_OutS( (const char*) buf, n);
    }

    virtual void ReadLine(BAS_STR& target)
    {
        // FIXME
    }

    virtual void BeginInput()
    {
        InputError = false;
        // Console:
        //   Read line into buffer. InputTo queries buffer.
        //   FIXME: should keep cursor on same line.
        ReadLine(InputLineBuffer);
        InputLineBufferPos=0;
    }

    virtual bool EndInputOkp()
    {
        // Console:
        //   Return false if InputTo failed.
        //   Return false if buffer is not empty.
        //   Return true otherwise.

        if(InputError) return false;
        IgnoreBlankInput();
        if(InputLineBufferPos < InputLineBuffer.size()) return false;
        return true;
    }

    virtual BAS_STR Input_N(size_t n)
    {
        // No echoing
        // FIXME
    }

    virtual int InputChar()
    {
        if(InputLineBufferPos < InputLineBuffer.size())
        {
            return InputLineBuffer[InputLineBufferPos++];
        }
        return -1;
    }

    virtual void GotOverflow()
    {
        InputError = true;
        *this << "Overflow\n";
    }
};

struct FileDescriptor
{
    QbFPstream fpout;
    QbFIstream fpin;
};
static std::map<int/*qbasic filenum*/, FileDescriptor> fdmap;

PrintStreamBase& GET_STREAM_OUTPUT(int filenumber)
{
    std::map<int, FileDescriptor>::iterator
        i = fdmap.find(filenumber);

    if(i == fdmap.end())
        B_ERROR(52); // bad file number
    
    return i->second.fpout;
}

InputStreamBase& GET_STREAM_INPUT(int filenumber)
{
    std::map<int, FileDescriptor>::iterator
        i = fdmap.find(filenumber);

    if(i == fdmap.end())
        B_ERROR(52); // bad file number
    
    return i->second.fpin;
}

static QbConsoleStream cons;

PrintStreamBase& GET_STDOUT()
{
    static bool tty_out = isatty(1);
    if(!tty_out) { static QbFPstream out(1); return out; }
    return cons;
}
InputStreamBase& GET_STDIN()
{
    static bool tty_in = isatty(0);
    if(!tty_in) { static QbFIstream in(0); return in; }
    return cons;
}
